dir=File.expand_path(File.dirname(__FILE__))
require File.join(dir, ".", "BTreeNode.rb")
require File.join(dir, ".", "Disk.rb")
require File.join(dir, ".", "Boolean.rb")



class BTree
  private
    attr_reader :disk,:root,:size,:root_address
  public
  
  def initialize(size=5)
    @size=size
    @disk=Disk.new()
    @root=BTreeNode.new(size,disk)
    @root.search(45)
    @root_address=disk.write(root)
  end
  
  def []=(key,value) 
    path=Array.new
    node,found=search(key, path)
    if(found) then
      node.set_value(key,value)
    else      
      add(key,value,path)
    end    
  end
  
  def [](key)
    current_node=root
    place,found=current_node.search(key)
    while(!found and !current_node.leaf?) do      
      current_node=disk.read(current_node.addresses[place])
      place,found=current_node.search(key)
    end	    
    current_node.values[place] if found
  end
  
  def invariant
    #every internal node has max m children
    
    #ever internal node exept root has at least (m/2).ceil children
    
    #root has at least 2 children exept when it is a leaf
    
    #internal node with k children has k-1 keys.
    
  end
  
  def each
    root.each{|node| yield node}
  end
  
  def add(key,value,path,next_address=0)    
    #puts "add #{key} => #{value} , #{next_address}"
    current_node=path.pop
    if(!current_node.full?) then
      #puts "node not full"
      current_node.add(key,value,next_address)
      if(path.empty?) then #root node
        #puts "add #{key} => #{value} , #{next_address} to root"
        disk.rewrite(root_address,root)
      else
        #puts "add #{key} => #{value} , #{next_address} to a regular node"
        position_node = path[-1]
        place,found=position_node.search(key)
        disk.rewrite(position_node.addresses[place],current_node)
      end
    else#node full
      if(path.empty?)#root node
        #puts "split root node"
        disk.rewrite(root_address,root)
        old_root_address=root_address
        old_root = current_node
        
        #new sibling node
        new_sibling,median_key,median_value = current_node.split(key,value,next_address)
        new_sibling_address=disk.write(new_sibling)
        
        #puts new_sibling.to_s + new_sibling_address.to_s
        #puts current_node.to_s + @root_address.to_s
        
        #new root node
        @root= BTreeNode.new(size,disk)
        @root_address=disk.write(root)
        @root.keys[0]=median_key
        @root.values[0]=median_value
        @root.addresses[0]=old_root_address
        @root.addresses[1]=new_sibling_address
        @root.leaf=false

        #save everything
        disk.rewrite(new_sibling_address,new_sibling)
        disk.rewrite(old_root_address,old_root)
        disk.rewrite(root_address,root)
      else
        #puts "split regular node"        
        new_sibling,median_key,median_value = current_node.split(key,value,next_address)
        new_sibling_address=disk.write(new_sibling)

        position_node = path[-1]
        place,found=position_node.search(key)
        
        disk.rewrite(position_node.addresses[place],current_node)
        disk.rewrite(new_sibling_address,new_sibling)
        
        #puts new_sibling.to_s + new_sibling_address.to_s
        #puts current_node.to_s + position_node.addresses[place].to_s
        
        add(median_key,median_value,path,new_sibling_address)      
      end
    end
  end
  
  def to_s
    contents="\n"
    root.each{ |node| contents << node.to_s }
    contents<<"\n"
  end
  
   
  protected
    def search(key,path)
      current_node=root
      path<<current_node
      current_node.search(key)
      place,found = current_node.search(key)
      while(!found and !current_node.leaf?)
        #puts current_node.leaf?
        current_node=disk.read(current_node.addresses[place])
        #puts current_node
        path<<current_node
        place,found=current_node.search(key)
      end
      [current_node,found]
    end
end

