dir=File.expand_path(File.dirname(__FILE__))
require File.join(dir, ".", "Boolean.rb")
require File.join(dir, ".", "RedBlackNode.rb")
require File.join(dir, ".", "Direction.rb")
require File.join(dir, ".", "BottomUp.rb")
require File.join(dir, ".", "TopDown.rb")
require File.join(dir, ".", "Color.rb")


class RedBlackTree  
  include BottomUp
  
  public
    def [](start_key,end_key=nil)
      if end_key then
        range_search(start_key,end_key)
      else
        search(start_key)
      end
    end
    
    
    def []=(key,data)
      place = find_place(key)[0]
      if place.root.nil? then
        add(key,data)
      else
        place.root.data=data
      end    
    end
    
    
    def to_s
      string = "nr\t-\tkey\t=>\tval\t-\tcolor \t-\tdepth\n"
      inorder = 0
      each do |node|
        inorder = inorder + 1
        string << "#{inorder}\t-\t#{node.to_s}\t-\t#{node.depth}\n"
      end
      string
    end
    
    #traverses the tree inorder
    def each
      return if root.nil?
      
      root.left.each{|node| yield node}
      yield root
      root.right.each{|node| yield node}
    end
    
    #traverses the tree from biggest to smallest value
    def each_biggest_first
      root.right.each_biggest_first{|node| yield node} unless root.right.root.nil?
      yield root
      root.left.each_biggest_first{|node| yield node} unless root.left.root.nil?      
    end
    
    def range_tree_search(upper_left,lower_right)
      points = Array.new
      
      split_node=split_node_search(upper_left,lower_right)
      
      return RedBlackTree.new if split_node.nil?
      
      current_node=split_node.left.root 
      while(current_node) do
        if upper_left <= current_node.key then
          puts "current node: #{current_node.key} #{lower_right.y <=current_node.key.y and current_node.key.y <= upper_left.y}"
          points << current_node.key if (upper_left <= current_node.key and current_node.key <= lower_right and lower_right.y <=current_node.key.y and current_node.key.y <= upper_left.y )
          if current_node.right.root
            current_node.right.root.data[lower_right.point_by_y,upper_left.point_by_y].each{|node|
            points << node.key 
            puts "1221" + node.key.to_s
            } 
          end
          puts current_node.right.root.data if current_node.right.root
          current_node=current_node.left.root              
        else
          current_node=current_node.right.root
        end
      end
      puts "split_node: #{split_node.key} #{lower_right.y <=split_node.key.y and split_node.key.y <= upper_left.y}"
points << split_node.key if (upper_left <= split_node.key and split_node.key <= lower_right and lower_right.y <=split_node.key.y and split_node.key.y <= upper_left.y )
      
      current_node=split_node.right.root 
      while(current_node) do
        if lower_right >= current_node.key then
          puts "current node: #{current_node.key} #{lower_right.y <=current_node.key.y and current_node.key.y <= upper_left.y}"
          points << current_node.key if (upper_left <= current_node.key and current_node.key <= lower_right and lower_right.y <=current_node.key.y and current_node.key.y <= upper_left.y )
          if current_node.left.root
          current_node.left.root.data[lower_right.point_by_y,upper_left.point_by_y].each do |node|
          points << node.key 
          puts lower_right.point_by_y.to_s + " " + upper_left.point_by_y.to_s + "dfsf " + node.key.to_s
          end
          end
          puts current_node.left.root.data if current_node.left.root
          current_node=current_node.right.root            
        else
          current_node=current_node.left.root
        end
      end
      points
    end
    
        def range_each(start_key,end_key)
      
      split_node=split_node_search(start_key,end_key)
      
      return RedBlackTree.new if split_node.nil?
      
      current_node=split_node.left.root 
      while(current_node) do
        if start_key <= current_node.key then
          yield current_node,false
          current_node.right.each { |node| yield node,true }
          current_node=current_node.left.root              
        else
          current_node=current_node.right.root
        end
      end
      
      yield split_node,false if(start_key <= split_node.key and split_node.key <= end_key  )
      
      current_node=split_node.right.root 
      while(current_node) do
        if end_key >= current_node.key then
          yield current_node,false
          current_node.left.each { |node| yield node,true }
          current_node=current_node.right.root            
        else
          current_node=current_node.left.root
        end
      end
    end

  
  protected
    attr_accessor  :root
    
    def range_search(start_key,end_key)
      nodes=Array.new
      range_each(start_key,end_key){ |node,in_subtree|  nodes << node   }
      nodes.sort{|n1,n2| n1.key<=>n2.key}
    end
    
    
    
    def split_node_search(start_key,end_key)
      current_node=root
      while(!current_node.nil? and (end_key <= current_node.key or start_key > current_node.key)) do
        if(end_key <=  current_node.key) then
          current_node=current_node.left.root
        else
          current_node=current_node.right.root
        end
      end
      current_node
    end
    
    def predecessor(node)
      under=node.left
      under=under.root.right while (! under.root.right.root.nil?)        
      under
    end
    
    #the color of nil nodes is black: reduces the number of special cases 
    def color
      return Color::BLACK if root.nil?
      root.color
    end 
    
    def red?
      color==Color::RED
    end
  
    def black?
      color==Color::BLACK
    end
    
    #checks if the tree is valid, raises errors if not
    def check()
      if(root.nil?)
        raise "nullpointer shoul be black" unless (black?)
        return
      end
      
      eve = root #eve=mother of all nodes
      eve = eve.parent while(!eve.parent.nil?)
      
      raise "Eve shoud be black: it began in africa ca ca...<#{eve.to_s}>" unless (eve.black?)
      raise "every red node should have two black children" if( root.red? and (root.left.red? or root.right.red?))
      raise "evert leaf should have two black children" if (root.left.root.nil? and root.right.root.nil? and (root.left.red? or root.right.red?))
      
      #check black depth
      black_depth()
      
      root.left.check()
      root.right.check()
    end
    
    #the number of black nodes on the way down
    def black_depth
      return 1 if root.nil?
      black_depth_left  = root.left.black_depth 
      black_depth_right = root.right.black_depth
      raise "black depth left not the same as right" unless (black_depth_left==black_depth_right)        
      return [black_depth_left+1,black_depth_left][color]        
    end
    
    #returns the data linked with key, returns nil if 
    #no node is found
    def search(key)
      place = find_place(key)[0]
      return nil if place.root.nil?
      place.root.data
    end
    
    #finds a place for a key 
    #returns an array with the place (tree) and the parent
    def find_place(key)
      place=self;
      parent=nil;
      while (!place.root.nil? and place.root.key!=key) do
        parent=place.root
        if (key < place.root.key) then
            place=place.root.left
        else
            place=place.root.right
        end
      end      
      return [place,parent]
    end
    
    def rotate(direction)
      #puts "rotate #{["right","left"][direction]}"
      up=nil;
      down=root;
      if (direction==Direction::LEFT) then
        up=root.right.root
        #raise "can not rotate this node (#{root.to_s}) to left" if (down.nil? or up.nil?)
        down.right.root=up.left.root        
        up.left.root.parent = down unless ( up.left.root.nil?)
        up.left.root=down
      else
        up=root.left.root
        #raise "can not rotate this node (#{root.to_s}) to right" if (down.nil? or up.nil?)
        down.left.root=up.right.root       
        up.right.root.parent=down unless (up.right.root.nil?)
        up.right.root=down
      end
      @root=up
      up.parent=down.parent
      down.parent=up
    end
    
    #checks if child is on the same side as its parent
    #returns true if the parent is the left child of the grandparent and the child the left child of the parent or
    #if the parent is the right child of the grandparent and the child the right child of the parent.
    #returns false otherwise.
    def on_same_side_as_parent(child)
      parent = child.parent
      grandparent = parent.parent
      
      right_exists = (!grandparent.right.root.nil? and !parent.right.root.nil?)   
      all_right = (right_exists and grandparent.right.root.key==parent.key and parent.right.root.key==child.key)
      return true if all_right
      
      left_exists = (!grandparent.left.root.nil? and !parent.left.root.nil?)
      all_left = (left_exists and grandparent.left.root.key==parent.key and parent.left.root.key==child.key)
      return all_left
    end
    
    #finds the tree that has node as root
    def find_tree(node)
      return self if (node.parent.nil?)
      return node.parent.left if (!node.parent.left.root.nil? and node.parent.left.root.key == node.key )
      return node.parent.right if (!node.parent.right.root.nil? and node.parent.right.root.key == node.key )
      raise "No tree for node: #{node}"
    end
end

#~ tree = RedBlackTree.new
#~ tree[15]=12.5
#~ tree[16]=12.5
#~ tree[12]=15
#~ tree[5]=12.5



#~ total_depth,number_of_nodes,reds,blacks = 0.0 , 0.0,0,0
#~ tree.each do |node|
  #~ total_depth =total_depth + node.depth
  #~ reds = reds + 1 if node.color == 1
  #~ blacks = blacks + 1 if node.color == 0
  #~ number_of_nodes = number_of_nodes + 1.0  
#~ end

#~ puts "Average depth: #{total_depth/number_of_nodes}"
#~ puts "Black nodes: #{blacks}"
#~ puts "Red nodes: #{reds}"
#~ puts tree.to_s 
#~ tree.delete(5)
#~ puts tree.to_s

